@@ -46,6 +46,9 @@ ALLEGRO_AUDIO_DRIVER *_al_kcm_driver = NULL;
 #if defined(ALLEGRO_CFG_KCM_PULSEAUDIO)
    extern struct ALLEGRO_AUDIO_DRIVER _al_kcm_pulseaudio_driver;
 #endif
+#if defined(ALLEGRO_SDL)
+   extern struct ALLEGRO_AUDIO_DRIVER _al_kcm_sdl_driver;
+#endif
 
 /* Channel configuration helpers */
 
@@ -168,9 +171,6 @@ static ALLEGRO_AUDIO_DRIVER_ENUM get_config_audio_driver(void)
    ALLEGRO_CONFIG *config = al_get_system_config();
    const char *value;
 
-   if (!config)
-      return ALLEGRO_AUDIO_DRIVER_AUTODETECT;
-
    value = al_get_config_value(config, "audio", "driver");
    if (!value || value[0] == '\0')
       return ALLEGRO_AUDIO_DRIVER_AUTODETECT;
@@ -196,6 +196,50 @@ static ALLEGRO_AUDIO_DRIVER_ENUM get_config_audio_driver(void)
    return ALLEGRO_AUDIO_DRIVER_AUTODETECT;
 }
 
+/* Function: al_get_num_audio_output_devices
+ */
+int al_get_num_audio_output_devices()
+{
+   if (_al_kcm_driver) {
+      if (_al_kcm_driver->get_output_devices) {
+         _AL_LIST* audio_devices = _al_kcm_driver->get_output_devices();
+         return _al_list_size(audio_devices);
+      }
+      else {
+         return -1;
+      }
+   }
+
+   return 0;
+}
+
+/* Function: al_get_audio_output_device
+ */
+const ALLEGRO_AUDIO_DEVICE* al_get_audio_output_device(int index)
+{
+   if (_al_kcm_driver) {
+      if (_al_kcm_driver->get_output_devices) {
+         _AL_LIST* audio_devices = _al_kcm_driver->get_output_devices();
+
+         if (index >= 0 && index < (int)_al_list_size(audio_devices)) {
+            return _al_list_item_data(_al_list_at(audio_devices, index));
+         }
+      }
+      else {
+         return NULL;
+      }
+   }
+
+   return NULL;
+}
+
+/* Function: al_get_audio_device_name
+ */
+const char* al_get_audio_device_name(const ALLEGRO_AUDIO_DEVICE * device)
+{
+   return device ? device->name : NULL;
+}
+
 static bool do_install_audio(ALLEGRO_AUDIO_DRIVER_ENUM mode)
 {
    bool retVal;
@@ -212,6 +256,11 @@ static bool do_install_audio(ALLEGRO_AUDIO_DRIVER_ENUM mode)
 
    switch (mode) {
       case ALLEGRO_AUDIO_DRIVER_AUTODETECT:
+#if defined(ALLEGRO_SDL)
+         retVal = do_install_audio(ALLEGRO_AUDIO_DRIVER_SDL);
+         if (retVal)
+            return retVal;
+#endif
 #if defined(ALLEGRO_CFG_KCM_AQUEUE)
          retVal = do_install_audio(ALLEGRO_AUDIO_DRIVER_AQUEUE);
          if (retVal)
@@ -349,6 +398,19 @@ static bool do_install_audio(ALLEGRO_AUDIO_DRIVER_ENUM mode)
             return false;
          #endif
 
+      case ALLEGRO_AUDIO_DRIVER_SDL:
+         #if defined(ALLEGRO_SDL)
+            if (_al_kcm_sdl_driver.open() == 0) {
+               ALLEGRO_INFO("Using SDL driver\n");
+               _al_kcm_driver = &_al_kcm_sdl_driver;
+               return true;
+            }
+            return false;
+         #else
+            _al_set_error(ALLEGRO_INVALID_PARAM, "SDL not available on this platform");
+            return false;
+         #endif
+
       default:
          _al_set_error(ALLEGRO_INVALID_PARAM, "Invalid audio driver");
          return false;
